import os
from collections import defaultdict
import platform
from jinja2 import Template
from pathlib import Path
from pycoze.reference.bot import ref_bot
from pycoze.reference.tool import ref_tools
from pycoze.reference.workflow import ref_workflow
import inspect


def function_to_string(func):
    # 获取函数签名
    signature = inspect.signature(func)

    # 获取函数名
    func_name = func.__name__

    # 获取函数的文档字符串
    docstring = func.__doc__

    # 格式化函数签名
    func_signature = f"def {func_name}{signature}:"

    # 构建最终的字符串
    result = f"#### {func_name.capitalize().replace('_', ' ')}\n"
    result += "**Function Signature:**\n"
    result += f"```python\n{func_signature}\n    \"\"\"{docstring}\n    \"\"\"\n    pass\n```\n\n"
    result += "**Usage:**\n"
    result += "```json\n"
    result += f"{{\n  \"{func_name}\": {{\n"

    # 添加参数
    for name, param in signature.parameters.items():
        result += f"    \"{name}\": \"{param.annotation.__name__ if param.annotation != inspect.Parameter.empty else 'value'}\"\n"

    result += "  }\n"
    result += "}\n"
    result += "```"

    return "\n\n" + result + "\n\n"

def get_abilities(bot_setting):
    abilities = []
    for bot_id in bot_setting["bots"]:
        bot = ref_bot(bot_id)
        if bot:
            abilities.append(bot)
    for tool_id in bot_setting["tools"]:
        abilities.extend(ref_tools(tool_id))
    for workflow_id in bot_setting["workflows"]:
        workflow = ref_workflow(workflow_id)
        if workflow:
            abilities.append(workflow)
    return abilities


def get_system_prompt(abilities, bot_setting):
    md_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), "prompt.md")
    with open(md_file, encoding="utf-8") as f:
        template = Template(f.read())

    if platform.system() == "Windows":
        cd_prompt = "When executing outside the working directory, include the CD command, such as cd /path/to/directory ; ls."
    else:
        cd_prompt = "When executing outside the working directory, include the CD command, such as cd /path/to/directory && ls."
    system = platform.system()


    abilities_str = "\n".join([function_to_string(a) for a in abilities])

    context = {
        "prompt": bot_setting["prompt"],
        "system": system,
        "cd_prompt": cd_prompt,
        "abilities_str": abilities_str,
        "programmer_mode": False,
        "no_exit_if_incomplete": False,
        "allow_read_file": False,
        "allow_read_multiple_files": False,
        "allow_execute_command": False,
        "allow_write_or_overwrite_file": False,
        "allow_replace_part_of_a_file": False,
        "allow_search_files": False,
        "allow_list_files": False,
        "allow_access_webpage": False,
    }

    context.update(bot_setting["systemAbility"])
    
    has_any_tool = False
    if len(abilities) > 0:
        has_any_tool = True
    for key in bot_setting["systemAbility"]:
        if key != "programmer_mode" and bot_setting["systemAbility"][key] == True:
            has_any_tool = True
            break
            
    context["has_any_tool"] = has_any_tool
    system_prompt = template.render(context)

    return system_prompt, has_any_tool


def resolve_relative_path(cwd:str, path_str: str) -> str:
    """返回基于CWD的规范化绝对路径"""
    path = Path(path_str)
    if path.is_absolute():
        return str(path.resolve())
    else:
        return str((Path(cwd) / path_str).resolve())


def read_local_file(file_path: str) -> str:
    """读取本地文件内容"""
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            return f.read()
    except:
        import chardet

        with open(file_path, "rb") as f:
            rawdata = f.read()
            encoding = chardet.detect(rawdata)["encoding"]
            return rawdata.decode(encoding)


# 全局忽略列表
IGNORE_LIST = [
    "node_modules",
    ".git",
    ".vscode",
    ".idea",
    "gitServer",
    ".DS_Store",
    "$RECYCLE.BIN",
    ".Trash-1000",
    ".Spotlight-V100",
    ".Trashes",
    ".TemporaryItems",
    ".fseventsd",
    "System Volume Information",
    "pycache",
    "env",
    "venv",
    "target/dependency",
    "build/dependencies",
    "dist",
    "out",
    "bundle",
    "vendor",
    "tmp",
    "temp",
    "deps",
    "pkg",
    "Pods",
]


def should_ignore(path):
    """检查路径是否在忽略列表中"""
    parts = path.split(os.sep)
    return any(part in IGNORE_LIST for part in parts)


def get_files_and_folders(root, recursive: bool):
    """递归获取所有文件和文件夹，并记录属性"""
    items = []

    # 使用 os.walk 遍历目录
    for dirpath, dirnames, filenames in os.walk(root):
        # 排除忽略列表中的路径
        if should_ignore(dirpath):
            continue

        # 记录文件夹
        for dirname in dirnames:
            full_path = os.path.join(dirpath, dirname)
            if not should_ignore(full_path):
                relative_path = os.path.relpath(full_path, root)
                is_empty = not os.listdir(full_path)
                depth = relative_path.count(os.sep)
                items.append(
                    (relative_path, "empty_folder" if is_empty else "folder", depth)
                )

        # 记录文件
        for filename in filenames:
            full_path = os.path.join(dirpath, filename)
            if not should_ignore(full_path):
                relative_path = os.path.relpath(full_path, root)
                depth = relative_path.count(os.sep)
                items.append((relative_path, "file", depth))

        # 如果 recursive 为 False，则只遍历当前目录，不进入子目录
        if not recursive:
            break

    return items


def format_filelist_str(items, limit):
    """根据limit格式化输出"""
    depth_groups = defaultdict(list)
    for item in items:
        depth_groups[item[2]].append(item)

    max_depth = max(depth_groups.keys(), default=0)
    show_list = []
    last_depth = 0

    # 浅层
    current_items = sorted(depth_groups[0], key=lambda x: x[0])
    overflow = len(current_items) > limit
    for item in current_items[:limit]:
        show_list.append(item)

    for depth in range(1, max_depth + 1):
        current_items = depth_groups[depth]
        if len(show_list) + len(current_items) <= limit:
            last_depth = depth
            for item in current_items:
                show_list.append(item)
        else:
            break

    result_str_list = []
    show_list.sort(key=lambda x: x[0])
    for item in show_list:
        if item[1] == "file":
            result_str_list.append(f"{item[0]}")
        elif item[1] == "folder" and item[2] == last_depth:
            result_str_list.append(f"{item[0]}/...more...")
        else:
            result_str_list.append(f"{item[0]}/")
    if overflow:
        result_str_list.append("...more...")

    return "\n".join(result_str_list)


def get_formatted_filelist_str(root: str, recursive: bool, limit=200):
    items = get_files_and_folders(root, recursive)
    return format_filelist_str(items, limit=limit)




import sys


def read_arg(param, is_path=False, posix=True):
    while param.startswith("-"):
        param = param[1:]
    args = sys.argv[1:]
    value = None
    for i in range(len(args)):
        if args[i] == "-" + param or args[i] == "--" + param:
            value = args[i + 1]
            break
    if is_path and value:
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        if posix:
            value = value.replace("\\", "/")

    return value





import time
from typing import Union
from pycoze import utils


socket = utils.socket


class ViewCls:

    def __init__(self, location: list[str], nodeIds: list[str] = []):
        self.location = location
        self.nodeIds = nodeIds
        self.is_in_workflow = len(nodeIds) > 0
        self.is_tab = len(nodeIds) == 0

    def __getitem__(self, key):
        return self.location[key]

    def wait_for_tab_open(self):
        times = 0
        while not self.is_tab_open():
            time.sleep(0.01)
            times += 1
            if times > 1000:
                raise Exception("Tab open timeout")

    def close_tab(self):
        assert self.is_tab
        self.wait_for_tab_open()
        socket.post("close-tab", {"location": self.location})

    def set_activate(self):
        assert self.is_tab
        self.wait_for_tab_open()
        socket.post("switchTab", {"location": self.location})

    def is_tab_open(self):
        assert self.is_tab
        result = socket.post_and_recv_result("is-tab-open", {"location": self.location})
        return result

    def pin_tab(self):
        assert self.is_tab
        self.wait_for_tab_open()
        socket.post("pin-tab", {"location": self.location})

    def unpin_tab(self):
        assert self.is_tab
        self.wait_for_tab_open()
        socket.post("unpin-tab", {"location": self.location})


class WorkflowCls(ViewCls):

    def layout_arrangement(self):
        # NetworkX layout arrangement，直接用python的布局即可，虽然没有输入输出port的，但是用节点坐标，再缩放几倍一般情况就可以了
        pass

    def run(self, wait_for_end=False):
        result = socket.post_and_recv_result(
            "workflow-run", {"location": self.location, "wait_for_end": wait_for_end}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]  # task_id

    def stop(self):
        result = socket.post_and_recv_result(
            "workflow-stop", {"location": self.location}
        )
        if not result["ok"]:
            raise Exception(result["value"])

    def get_state(self):
        result = socket.post_and_recv_result(
            "workflow-get-state", {"location": self.location}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_nodes(self):
        result = socket.post_and_recv_result(
            "workflow-get-graph-nodes", {"location": self.location}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_edges(self):
        result = socket.post_and_recv_result(
            "workflow-get-graph-edges", {"location": self.location}
        )
        return result["value"]

    def add_text_node(self, text: str, position="center"):
        result = socket.post_and_recv_result(
            "workflow-add-text-node",
            {"location": self.location, "text": text, "position": position},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        time.sleep(0.01)
        return result["value"]  # node

    def add_markdown_node(self, text: str, position="center"):
        result = socket.post_and_recv_result(
            "workflow-add-markdown-node",
            {"location": self.location, "text": text, "position": position},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        time.sleep(0.01)
        return result["value"]  # node

    def add_file_node(self, file_path: str, position="center"):
        is_step = False
        if file_path.lower().endswith(".py"):
            with open(file_path, "r", encoding="utf-8") as f:
                content = f.read()
                if "ps_workflow" in content:
                    is_step = True
        if is_step:
            return self.add_step_file_node(file_path, position)
        result = socket.post_and_recv_result(
            "workflow-add-file-node",
            {"location": self.location, "file_path": file_path, "position": position},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        time.sleep(0.01)
        return result["value"]  # node

    def add_step_file_node(self, file_path: str, position="center"):
        result = socket.post_and_recv_result(
            "workflow-add-step-file-node",
            {"location": self.location, "file_path": file_path, "position": position},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        time.sleep(0.01)
        return result["value"]  # node

    def add_edge(
        self, from_node_id: str, from_port_id: str, to_node_id: str, to_port_id: str
    ):
        result = socket.post_and_recv_result(
            "workflow-add-edge",
            {
                "location": self.location,
                "source": {"cell": from_node_id, "port": from_port_id},
                "target": {"cell": to_node_id, "port": to_port_id},
            },
        )
        if not result["ok"]:
            raise Exception(result["value"])
        time.sleep(0.01)
        return result["value"]  # edge

    def get_node_info(self, node: Union[str, object]):
        if isinstance(node, object):
            node_id = node["id"]
        result = socket.post_and_recv_result(
            "workflow-get-node-info", {"location": self.location, "node_id": node_id}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_node_state(self, node: Union[str, object]):
        if isinstance(node, object):
            node_id = node["id"]
        result = socket.post_and_recv_result(
            "workflow-get-node-state", {"location": self.location, "node_id": node_id}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_node_log(self, node: Union[str, object]):
        if isinstance(node, object):
            node_id = node["id"]
        result = socket.post_and_recv_result(
            "workflow-get-node-log", {"location": self.location, "node_id": node_id}
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_node_input_data(self, node: Union[str, object], port: str):
        if isinstance(node, object):
            node_id = node["id"]
        result = socket.post_and_recv_result(
            "workflow-get-node-input-data",
            {"location": self.location, "node_id": node_id, "port": port},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def get_node_output_data(self, node: Union[str, object], port: str):
        if isinstance(node, object):
            node_id = node["id"]
        result = socket.post_and_recv_result(
            "workflow-get-node-output-data",
            {"location": self.location, "node_id": node_id, "port": port},
        )
        if not result["ok"]:
            raise Exception(result["value"])
        return result["value"]

    def parse_node_data(self, data: Union[dict, list]):
        if isinstance(data, list):
            return [parse_data(d) for d in data]
        return parse_data(data)



from .arg import read_arg
from .env import read_params_file, params, read_json_file
from .socket import TcpSocket, socket, socket_subscribe
from .text_or_file import to_text

__all__ = [read_arg, 
           read_params_file, params, read_json_file, 
           TcpSocket, socket, socket_subscribe,
           to_text]
